# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dict_typer']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.1.1,<8.0.0', 'rich>=1.0.0,<2.0.0']

entry_points = \
{'console_scripts': ['dict-typer = dict_typer:cli']}

setup_kwargs = {
    'name': 'dict-typer',
    'version': '0.1.10',
    'description': 'A simple library to take data and turn it into python typing definitions.',
    'long_description': '# Dict-Typer\n\nA simple tool to take a json payload and convert it into Python TypedDict class\ndefinitions\n\n## Why is this useful?\n\nWhen dealing with API responses, you\'re very likely to be using JSON responses,\nand you might have deeply nested dictionaries, lists of items and it can be\nslightly hard to wrap your head around the structure of the responess you are\nworking with. The first thing I usually do it try to create some data structure\naround it so that I can benefit from linting and typing information in my code.\n\nNow this tends to be time consuming and error prone, so I thought it might be a\ngood idea to automate this process with a tool for the future. Just as an\nexample, if we take the output generated from the Example section below and\nimagine it\'s a response we get from an api. We can plug it in like this:\n\n```python\nfrom project.types import Root\n\n\ndef get_from_api() -> Root:\n    pass\n\n\ndef run() -> None:\n    response = get_from_api()\n\n    test1 = response["nested_dict"]["number"] + 1\n    test2 = response["nested_dict"]["string"] + 1\n    test3 = response["nested_dict"]["non_existant"] + 1\n    for item in response["optional_items"]:\n        print(item + 1)\n```\n\nand if we run mypy on this\n\n```shell\n-> % poetry run mypy test.py\ntest.py:43: error: Unsupported operand types for + ("str" and "int")\ntest.py:44: error: TypedDict "NestedDict" has no key \'non_existant\'\ntest.py:46: error: Unsupported operand types for + ("None" and "int")\ntest.py:46: error: Unsupported operand types for + ("str" and "int")\ntest.py:46: note: Left operand is of type "Union[None, int, str]"\nFound 4 errors in 1 file (checked 1 source file)\n```\n\nit will immediately detect four issues!\n\nI also want to use this project to learn more about analysing code, making sure\nthe project is well tested so that it\'s easy to experiment and try different\napproaches.\n\n## Usage\n\nEither supply a path to a file or pipe json output to `dict-typer`\n\n```help\n-> % dict-typer --help\n\nUsage: dict-typer [OPTIONS] [FILE]...\n\nOptions:\n  --imports / --no-imports  Show imports at the top, default: True\n  -r, --rich                Show rich output.\n  -l, --line-numbers        Show line numbers if rich.\n  --version                 Show the version and exit.\n  --help                    Show this message and exit.\n\n-> % dict-typer ./.example.json\n...\n\n-> % curl example.com/test.json | dict-typer\n...\n```\n\n## TypeDict definitions\n\nThere are two ways to define a TypedDict, the primary one that uses the class\nbased structure, as seen in the examples here. It\'s easier to read, but it has\na limitation that the each key has to be avalid identifier and not a reserved\nkeyword. Normally that\'s not an issue, but if you have for example, the\nfollowing data\n\n```json\n{\n    "numeric-id": 123,\n    "from": "far away",\n}\n```\n\nwhich is valid json, but has the invalid identifier `numeric-id` and reserved\nkeyword `from`, meaning the definition\n\n```python\nclass Root(TypedDict):\n    numeric-id: int\n    from: str\n```\n\nis invalid. In detected cases, dict-typer will use an [alternative\nway](https://www.python.org/dev/peps/pep-0589/#alternative-syntax) to define\nthose types, looking like this\n\n```python\nRoot = TypedDict(\'Root\', {\'numeric-id\': int, \'from\': str\'})\n```\n\nwhich is not as readable, but valid.\n\ndict-typer by default only uses the alternative definition for the types with\ninvalid keys.\n\n## Lists\n\nIf the root of the payload is a list, it will be treated just like a list\nwithin a dictionary, where each item is parsed and definitions created for sub\nitems. In these cases, a type alias is added as well to the output to capture\nthe type of the list. For example, the list `[1, "2", 3.0, { "id": 123 }, {\n"id": 456 }]` will generate the following definitions:\n\n```python\nfrom typing_extensions import TypedDict\n\n\nclass RootItem(TypedDict):\n    id: int\n\nRoot = List[Union[RootItem, float, int, str]]\n```\n\n## Examples\n\n### Calling from shell\n\n```shell\n-> % cat .example.json\n{\n  "number_int": 123,\n  "number_float": 3.0,\n  "string": "string",\n  "list_single_type": ["a", "b", "c"],\n  "list_mixed_type": ["1", 2, 3.0],\n  "nested_dict": {\n    "number": 1,\n    "string": "value"\n  },\n  "same_nested_dict": {\n    "number": 2,\n    "string": "different value"\n  },\n  "multipe_levels": {\n    "level2": {\n      "level3": {\n        "number": 3,\n        "string": "more values"\n      }\n    }\n  },\n  "nested_invalid": { "numeric-id": 123, "from": "far away" },\n  "optional_items": [1, 2, "3", "4", null, 5, 6, null]\n}\n\n-> % cat .example.json | dict-typer\nfrom typing import List, Union\n\nfrom typing_extensions import TypedDict\n\n\nclass NestedDict(TypedDict):\n    number: int\n    string: str\n\n\nclass Level2(TypedDict):\n    level3: NestedDict\n\n\nclass MultipeLevels(TypedDict):\n    level2: Level2\n\n\nNestedInvalid = TypedDict("NestedInvalid", {\n    "numeric-id": int,\n    "from": str,\n})\n\n\nclass Root(TypedDict):\n    number_int: int\n    number_float: float\n    string: str\n    list_single_type: List[str]\n    list_mixed_type: List[Union[float, int, str]]\n    nested_dict: NestedDict\n    same_nested_dict: NestedDict\n    multipe_levels: MultipeLevels\n    nested_invalid: NestedInvalid\n    optional_items: List[Union[None, int, str]]\n```\n\n### Calling from Python\n```python\nIn [1]: source = {\n   ...:   "number_int": 123,\n   ...:   "number_float": 3.0,\n   ...:   "string": "string",\n   ...:   "list_single_type": ["a", "b", "c"],\n   ...:   "list_mixed_type": ["1", 2, 3.0],\n   ...:   "nested_dict": {\n   ...:     "number": 1,\n   ...:     "string": "value"\n   ...:   },\n   ...:   "same_nested_dict": {\n   ...:     "number": 2,\n   ...:     "string": "different value"\n   ...:   },\n   ...:   "multipe_levels": {\n   ...:     "level2": {\n   ...:       "level3": {\n   ...:         "number": 3,\n   ...:         "string": "more values"\n   ...:       }\n   ...:     }\n   ...:   },\n   ...:   "nested_invalid": { "numeric-id": 123, "from": "far away" },\n   ...:   "optional_items": [1, 2, "3", "4", None, 5, 6, None]\n   ...: }\n   ...:\n\nIn [2]: from dict_typer import get_type_definitions\n\nIn [3]: print(get_type_definitions(source, show_imports=True))\nfrom typing import List, Union\n\nfrom typing_extensions import TypedDict\n\n\nclass NestedDict(TypedDict):\n    number: int\n    string: str\n\n\nclass Level2(TypedDict):\n    level3: NestedDict\n\n\nclass MultipeLevels(TypedDict):\n    level2: Level2\n\n\nNestedInvalid = TypedDict("NestedInvalid", {\n    "numeric-id": int,\n    "from": str,\n})\n\n\nclass Root(TypedDict):\n    number_int: int\n    number_float: float\n    string: str\n    list_single_type: List[str]\n    list_mixed_type: List[Union[float, int, str]]\n    nested_dict: NestedDict\n    same_nested_dict: NestedDict\n    multipe_levels: MultipeLevels\n    nested_invalid: NestedInvalid\n    optional_items: List[Union[None, int, str]]\n```\n',
    'author': 'Axel',
    'author_email': 'dict.typer@absalon.is',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ikornaselur/dict-typer',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from typing import Any, Dict, List, Optional, Set, Tuple, Type, Union

from dict_typer.exceptions import ConvertException
from dict_typer.models import (
    DictEntry,
    EntryType,
    MemberEntry,
    key_to_dependency_cmp,
    sub_members_to_imports,
    sub_members_to_string,
)
from dict_typer.utils import key_to_class_name

BASE_TYPES: Tuple[Type, ...] = (  # type: ignore
    bool,
    bytearray,
    bytes,
    complex,
    enumerate,
    float,
    int,
    memoryview,
    range,
    str,
    type,
    filter,
    map,
    zip,
)


class DefinitionBuilder:
    definitions: List[DictEntry]
    root_list: Set[MemberEntry]
    root_type_name: str
    type_postfix: str
    show_imports: bool
    source: Union[Dict, List]

    _output: Optional[str] = None

    def __init__(
        self,
        source: Union[Dict, List],
        root_type_name: str = "Root",
        type_postfix: str = "",
        show_imports: bool = True,
    ) -> None:
        self.definitions = []
        self.root_list = set()

        self.root_type_name = root_type_name
        self.type_postfix = type_postfix
        self.show_imports = show_imports

        self.source = source

    def _add_definition(self, entry: EntryType) -> EntryType:
        """ Add an entry to the definions.

        If the entry is a DictEntry and there's an existing entry with the same
        keys, then combine the DictEntries
        """
        if isinstance(entry, MemberEntry):
            self.root_list.add(entry)
            return entry
        else:
            dicts_only = [d for d in self.definitions if isinstance(d, DictEntry)]
            for definition in dicts_only:
                if entry.keys == definition.keys:
                    definition.update_members(entry.members)
                    return definition
                if entry.name == definition.name:
                    idx = 1
                    new_name = f"{entry.name}{idx}"
                    dicts_names = [d.name for d in dicts_only]
                    while new_name in dicts_names:
                        idx += 1
                        new_name = f"{entry.name}{idx}"
                    entry.name = new_name
            self.definitions.append(entry)
            return entry

    def _convert_list(self, key: str, lst: List, item_name: str) -> MemberEntry:
        entry = MemberEntry(key)

        idx = 0
        for item in lst:
            item_type = self._get_type(item, key=f"{item_name}{idx}")

            entry.sub_members.add(item_type)
            if isinstance(item_type, DictEntry):
                self._add_definition(item_type)
            idx += 1

        return entry

    def _convert_dict(self, type_name: str, dct: Dict) -> DictEntry:
        entry = DictEntry(type_name)
        for key, value in dct.items():
            value_type = self._get_type(value, key=key)
            if isinstance(value_type, DictEntry):
                definition = self._add_definition(value_type)
                value_type = definition
            entry.members[key] = {value_type}

        return entry

    def _get_type(self, item: Any, key: str) -> Union[MemberEntry, DictEntry]:
        if item is None:
            return MemberEntry("None")

        if isinstance(item, BASE_TYPES):
            return MemberEntry(type(item).__name__)

        if isinstance(item, (list, set, tuple, frozenset)):
            if isinstance(item, list):
                sequence_type_name = "List"
            elif isinstance(item, set):
                sequence_type_name = "Set"
            elif isinstance(item, frozenset):
                sequence_type_name = "FrozenSet"
            else:
                sequence_type_name = "Tuple"

            list_item_types: Set[Union[MemberEntry, DictEntry]] = set()
            idx = 0
            for value in item:
                item_type = self._get_type(value, key=f"{key}Item{idx}")
                if isinstance(item_type, DictEntry):
                    if item_type.members not in (
                        getattr(t, "members", None) for t in list_item_types
                    ):
                        list_item_types.add(item_type)
                        idx += 1
                        if isinstance(item_type, DictEntry):
                            self._add_definition(item_type)
                else:
                    list_item_types.add(item_type)

            return MemberEntry(sequence_type_name, sub_members=list_item_types)

        if isinstance(item, dict):
            if item == {}:
                return MemberEntry("Dict")
            return self._convert_dict(
                f"{key_to_class_name(key)}{self.type_postfix}", item
            )

        raise NotImplementedError(f"Type handling for '{type(item)}' not implemented")

    def build_output(self) -> str:
        if self._output:
            return self._output

        # Create the definitions from the source
        if isinstance(self.source, dict):
            self._add_definition(
                self._convert_dict(
                    f"{self.root_type_name}{self.type_postfix}", self.source
                )
            )
        else:
            self._add_definition(
                self._convert_list(
                    f"List", self.source, item_name=f"{self.root_type_name}Item"
                )
            )

        # Convert the definitions to structured output
        self._output = ""

        if self.show_imports:
            typing_imports = set()
            typed_dict_import = False

            for definition in self.definitions:
                if isinstance(definition, DictEntry) and definition.members:
                    typed_dict_import = True
                typing_imports |= definition.get_imports()
            if self.root_list:
                typing_imports.add("List")
                typing_imports |= sub_members_to_imports(self.root_list)

            if typing_imports:
                self._output += "\n".join(
                    [f"from typing import {', '.join(sorted(typing_imports))}", "", ""]
                )
            if typed_dict_import:
                self._output += "\n".join(
                    ["from typing_extensions import TypedDict", "", "", ""]
                )

        self._output += "\n\n\n".join(
            [str(d) for d in sorted(self.definitions, key=key_to_dependency_cmp)]
        )

        if self.root_list:
            if len(self._output):
                self._output += "\n"
                if len(self.definitions):
                    self._output += "\n\n"
            self._output += f"{self.root_type_name}{self.type_postfix} = {sub_members_to_string(self.root_list)}"
        # Special case with root being empty Dict
        if self.source == {}:
            if len(self._output):
                self._output += "\n"
            self._output += f"{self.root_type_name}{self.type_postfix} = Dict"

        return self._output


def get_type_definitions(
    source: Union[Dict, List],
    root_type_name: str = "Root",
    type_postfix: str = "",
    show_imports: bool = True,
) -> str:
    if not isinstance(source, (list, dict)):
        raise ConvertException(f"Unsupported source type: {type(source)}")

    builder = DefinitionBuilder(
        source,
        root_type_name=root_type_name,
        type_postfix=type_postfix,
        show_imports=show_imports,
    )

    return builder.build_output()



import io
import json
import sys
from typing import Dict, Tuple

import click
from rich.console import Console
from rich.syntax import Syntax

from dict_typer.type_definitions import get_type_definitions

__version__ = "0.1.10"


@click.command()
@click.option(
    "--imports/--no-imports",
    default=True,
    help="Show imports at the top, default: True",
)
@click.option("--rich", "-r", is_flag=True, help="Show rich output.")
@click.option(
    "--line-numbers", "-l", is_flag=True, help="Show line numbers if rich.",
)
@click.argument("file", type=click.File("r"), nargs=-1)
@click.version_option(__version__)
def cli(
    file: Tuple[io.TextIOWrapper],
    imports: bool = True,
    rich: bool = False,
    line_numbers: bool = False,
) -> None:
    if len(file) > 1:
        raise click.BadArgumentUsage("Multiple files supplied, run with one at a time")
    if len(file) == 0:
        if sys.stdin.isatty():
            raise click.UsageError(
                "Either provide the path to the file or pipe a file to dict-typer"
            )
        else:
            stream = sys.stdin.read().strip()
    else:
        stream = file[0].read().strip()

    try:
        parsed: Dict = json.loads(stream)
    except json.decoder.JSONDecodeError as e:
        raise click.UsageError(f"JSON serialisation error \n\n{e}")

    output = get_type_definitions(parsed, show_imports=imports)
    if rich:
        syntax = Syntax(output, "python", theme="monokai", line_numbers=line_numbers)
        console = Console()
        console.print(syntax)
    else:
        click.echo(output)
